from pyspark import SparkConf
from pyspark.sql import SparkSession
from pyspark.sql.functions import pandas_udf, PandasUDFType
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from sklearn import preprocessing

# This example is to show how to use the pandas udf in pyspark program

feature_list = ['v1','v2','v3']


# define the pandas udf
@pandas_udf("id long, pc1 double, pc2 double, prediction int", PandasUDFType.GROUPED_MAP)
def normalize(pdf):
    pdf_scaled = preprocessing.scale(pdf[feature_list])
    pdf = pdf.assign(v1=pdf_scaled[:,0],
                     v2=pdf_scaled[:,1],
                     v3=pdf_scaled[:,2])
    pca = PCA(n_components=2)

    X = pca.fit_transform(pdf[feature_list])
    pdf = pdf.assign(pc1=X[:,0],
                     pc2=X[:,1])

    kmeans = KMeans(n_clusters=3).fit(pdf[['pc1', 'pc2']])
    pdf = pdf.assign(prediction=kmeans.labels_)

    return pdf[['id', 'pc1', 'pc2', 'prediction']]


if __name__ == "__main__":
    spark = SparkSession.builder.config(conf=SparkConf()).getOrCreate()
    spark.conf.set("spark.sql.execution.arrow.enabled", "true")

    df = spark.createDataFrame(
        [
            (3, 1.0, 4.0, 57.0, 4.0, "a"),
            (1, 15.0, 19.0, 4.0, 3.0, "a"),
            (1, 21.0, 31.0, 54.0, 3.0, "a"),
            (1, 1.0, 3.0, 11.0, 4.0, "a"),
            (1, 1.0, 87.0, 1.0, 4.0, "a"),
            (1, 12.0, 13.0, 46.0, 63.0, "a"),
            (1, 1.0, 46.0, 52.0, 3.0, "a"),
            (2, 11.0, 45.0, 3.0, 5.0, 'c'),
            (2, 12.0, 34.0, 56.0, 73.0, 'b'),
            (2, 13.0, 4.0, 3.0, 5.0, 'c'),
            (2, 1.0, 4.0, 5.0, 7.0, 'b'),
            (3, 12.0, 42.0, 33.0, 56.0, 'c'),
            (3, 12.0, 4.0, 5.0, 7.0, 'b'),
            (3, 1.0, 4.0, 3.0, 5.0, 'c'),
            (2, 25.0, 72.0, 12.0, 42.0, 'b'),
            (2, 10.0, 33.0, 15.0, 4.0, "a")
        ],
        ("id", "v1", "v2", "v3", "v4", "type"))

    # use the pandas udf
    res = df.groupby("id").apply(normalize)
    res.printSchema()

    res.show(truncate=False)
    res.write.josn("obs://bucket/xxx/xx/")

    spark.stop()
    # expected result is like:
    # root
    #  |-- id: long (nullable = true)
    #  |-- pc1: double (nullable = true)
    #  |-- pc2: double (nullable = true)
    #  |-- prediction: integer (nullable = true)
