package com.huaweicloud.dli.io.common;

import com.obs.services.ObsClient;
import com.obs.services.exception.ObsException;
import com.obs.services.model.AppendObjectRequest;
import com.obs.services.model.AppendObjectResult;
import com.obs.services.model.GetObjectRequest;
import com.obs.services.model.ListObjectsRequest;
import com.obs.services.model.ObjectListing;
import com.obs.services.model.ObsObject;

import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import lombok.extern.slf4j.Slf4j;

@Slf4j
public class ObsProxy {
    private final ObsClient client;

    public ObsProxy(ObsClient client) {
        this.client = client;
    }

    public String getBucketName(String fullPath) {
        try {
            URI url = new URI(fullPath);
            return url.getHost();
        } catch (URISyntaxException e) {
            throw new RuntimeException("Failed to get bucket name from full path", e);
        }
    }

    public String getObjectPrefix(String fullPath) {
        try {
            URI url = new URI(fullPath);
            return url.getPath().substring(1);
        } catch (URISyntaxException e) {
            throw new RuntimeException("Failed to get object key from full path", e);
        }
    }

    public List<ObsObject> listObjects(String bucketName, String objPrefix) {
        ListObjectsRequest listObjReq = new ListObjectsRequest();
        listObjReq.setBucketName(bucketName);
        listObjReq.setPrefix(objPrefix);
        listObjReq.setMaxKeys(1000);
        ObjectListing objListing;
        List<ObsObject> obsObjList = new ArrayList<>();
        long startTime = System.currentTimeMillis();
        do {
            objListing = client.listObjects(listObjReq);
            obsObjList.addAll(objListing.getObjects().stream()
                .filter(object -> object.getMetadata().getContentLength() > 0)
                .collect(Collectors.toList()));
            listObjReq.setMarker(objListing.getNextMarker());
        } while (objListing.isTruncated());
        long endTime = System.currentTimeMillis();
        log.info("Get objects total cost: {}ms", endTime - startTime);

        return obsObjList;
    }

    public InputStream getObjectStream(String bucketName, String objKey) {
        GetObjectRequest getObjReq = new GetObjectRequest(bucketName, objKey);
        try {
            return client.getObject(getObjReq).getObjectContent();
        } catch (ObsException e) {
            if (e.getResponseCode() == 404) {
                log.warn("The obs object({}) to get is not fount", objKey);
                return null;
            } else {
                throw new RuntimeException("Failed to get obs object(" + objKey + ") content", e);
            }
        }
    }

    public AppendObjectResult appendObject(String bucketName, String objKey, InputStream inputStream) {
        return appendObject(bucketName, objKey, -1, inputStream);
    }

    public AppendObjectResult appendObject(String bucketName, String objKey, long position, InputStream inputStream) {
        AppendObjectRequest appendObjReq = new AppendObjectRequest();
        appendObjReq.setBucketName(bucketName);
        appendObjReq.setObjectKey(objKey);
        if (position != -1) {
            appendObjReq.setPosition(position);
        }
        appendObjReq.setInput(inputStream);
        return client.appendObject(appendObjReq);
    }
}
