package com.huaweicloud.dli.io.read;

import com.huaweicloud.dli.io.common.Column;
import com.huaweicloud.dli.io.common.DataType;
import com.huaweicloud.dli.io.common.ObsProxy;
import com.huaweicloud.dli.io.common.Row;
import com.huaweicloud.sdk.core.utils.JsonUtils;
import com.obs.services.model.ObsObject;
import com.univocity.parsers.common.ParsingContext;
import com.univocity.parsers.common.ResultIterator;
import com.univocity.parsers.csv.CsvParser;
import com.univocity.parsers.csv.CsvParserSettings;
import com.univocity.parsers.csv.UnescapedQuoteHandling;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.nio.charset.StandardCharsets;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Base64;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import lombok.extern.slf4j.Slf4j;

@Slf4j
public class CsvObsReader implements Reader {
    private static final String NULL_VALUE_FOR_CSV = "\u00A0\u0008\uFEFF";

    private static final int MAX_RECONNECT_NUM = 5;

    private static final ThreadLocal<DateFormat> DATE_FORMAT =
        ThreadLocal.withInitial(() -> new SimpleDateFormat("yyyy-MM-dd"));

    private final ObsProxy obsProxy;

    private final List<Column> resultSchema;

    private final List<ObsObject> objects;

    private int totalIter = 0;

    private int nextIterIdx = 0;

    private CsvParserSettings parserSettings;

    private ResultIterator<String[], ParsingContext> iterator;

    private int readRecordsPerIter = 0;

    private boolean closed = false;

    public CsvObsReader(ObsProxy obsProxy, List<ObsObject> objects, List<Column> resultSchema) {
        this.obsProxy = obsProxy;
        this.objects = objects;
        this.resultSchema = resultSchema;
    }

    @Override
    public void openReader() throws IOException {
        totalIter = objects.size();
        log.info("object size: {}", totalIter);

        parserSettings = new CsvParserSettings();
        parserSettings.getFormat().setDelimiter(',');
        parserSettings.getFormat().setQuote('\"');
        parserSettings.getFormat().setQuoteEscape('\\');
        parserSettings.getFormat().setComment('\u0000');
        parserSettings.getFormat().setLineSeparator("\n");
        parserSettings.setIgnoreLeadingWhitespaces(false);
        parserSettings.setIgnoreTrailingWhitespaces(false);
        parserSettings.setReadInputOnSeparateThread(false);
        parserSettings.setSkipEmptyLines(true);
        parserSettings.setNullValue(NULL_VALUE_FOR_CSV);
        parserSettings.setEmptyValue("");
        parserSettings.setMaxCharsPerColumn(-1);
        parserSettings.setUnescapedQuoteHandling(UnescapedQuoteHandling.STOP_AT_DELIMITER);
        iterator = getNextIterator();
    }

    private ResultIterator<String[], ParsingContext> getNextIterator() {
        if (totalIter > 0 && nextIterIdx < totalIter) {
            readRecordsPerIter = 0;
            return createIterator(false);
        } else {
            return null;
        }
    }

    private ResultIterator<String[], ParsingContext> createIterator(boolean recreate) {
        int currIterIdx = recreate ? nextIterIdx - 1 : nextIterIdx++;
        ObsObject object = objects.get(currIterIdx);
        log.info("{} iterator for object: {}", (recreate ? "recreate" : "create"), object.getObjectKey());
        InputStream inputStream = obsProxy.getObjectStream(object.getBucketName(), object.getObjectKey());
        return new CsvParser(parserSettings).iterate(inputStream, StandardCharsets.UTF_8).iterator();
    }

    @Override
    public Row read() throws IOException {
        if (closed) {
            throw new IllegalStateException("The reader state is closed");
        }

        if (!iterator.hasNext()) {
            log.info("read {} records from pre iterator", readRecordsPerIter);
            if (nextIterIdx < totalIter) {
                iterator = getNextIterator();
            }
        }
        if (iterator != null && iterator.hasNext()) {
            String[] values = readWithRetry();
            Row row = new Row(resultSchema);
            updateRow(values, row);
            return row;
        } else {
            return null;
        }
    }

    private String[] readWithRetry() throws IOException {
        String[] values;
        for (int times = 0; times < MAX_RECONNECT_NUM; times++) {
            try {
                values = iterator.next();
                readRecordsPerIter++;
                return values;
            } catch (Exception e) {
                log.warn("get record failed, may be connection closed by obs, so reconnect it!!!", e);
                reconnectAndSkipHandled();
            }
        }
        throw new IOException("Failed to read value with retry");
    }

    private void reconnectAndSkipHandled() throws IOException {
        iterator = createIterator(true);
        log.warn("reconnect, and skip records: {}", readRecordsPerIter);
        // skip all records have read
        long startTime = System.currentTimeMillis();
        for (int i = 0; i < readRecordsPerIter; i++) {
            if (iterator.hasNext()) {
                iterator.next();
            } else {
                throw new IOException(String.format(
                    Locale.ROOT,
                    "Unexpected error, pre read records: %d, but now only read records: %d",
                    readRecordsPerIter, (i + 1)));
            }
        }
        long endTime = System.currentTimeMillis();
        log.info("skip records cost: {}ms", (endTime - startTime));
    }

    private void updateRow(String[] values, Row row) throws IOException {
        if (values.length != row.getSchema().size()) {
            String msg = String.format("Data value size: %s not match column size: %s",
                values.length, row.getSchema().size());
            log.error(msg + ", and data after parse details:");
            for (String data : values) {
                log.error("\t{}", data);
            }
            throw new IOException(msg);
        }
        List<Column> columns = row.getSchema();
        for (int i = 0; i < columns.size(); i++) {
            Column column = columns.get(i);
            updateCol(row, i, column.getType(), values[i]);
        }
    }

    @SuppressWarnings({"unchecked"})
    private void updateCol(Row row, int i, DataType dataType, String data) throws IOException {
        DataType.TypeName typeName = DataType.TypeName.fromName(dataType.getName());
        switch (typeName) {
            case BOOLEAN:
                Boolean boolData = isNull(data) ? null : Boolean.valueOf(data);
                row.setBoolean(i, boolData);
                break;
            case TINYINT:
                Byte byteData = isNull(data) ? null : Byte.valueOf(data);
                row.setTinyint(i, byteData);
                break;
            case SMALLINT:
                Short shortData = isNull(data) ? null : Short.valueOf(data);
                row.setSmallint(i, shortData);
                break;
            case INT:
                Integer intData = isNull(data) ? null : Integer.valueOf(data);
                row.setInt(i, intData);
                break;
            case BIGINT:
                Long longData = isNull(data) ? null : Long.valueOf(data);
                row.setBigint(i, longData);
                break;
            case FLOAT:
                Float floatData = isNull(data) ? null : Float.valueOf(data);
                row.setFloat(i, floatData);
                break;
            case DOUBLE:
                Double doubleData = isNull(data) ? null : Double.valueOf(data);
                row.setDouble(i, doubleData);
                break;
            case DECIMAL:
                BigDecimal decimalData = isNull(data) ? null : new BigDecimal(data);
                row.setDecimal(i, decimalData);
                break;
            case STRING:
                row.setString(i, isNull(data) ? null : data);
                break;
            case DATE:
                try {
                    Date dateData = isNull(data) ? null : DATE_FORMAT.get().parse(data);
                    row.setDate(i, dateData);
                } catch (ParseException e) {
                    log.error("Failed to parse date data: {}", data, e);
                    throw new IOException("Failed to parse date data: " + data);
                }
                break;
            case TIMESTAMP:
                Timestamp tsData = isNull(data) ? null : new Timestamp(Long.parseLong(data));
                row.setTimestamp(i, tsData);
                break;
            case BINARY:
                byte[] bytes = isNull(data) ? null : Base64.getDecoder().decode(data);
                row.setBinary(i, bytes);
                break;
            case ARRAY:
                List<Object> arrData = isNull(data) ? null : JsonUtils.toObject(data, List.class);
                row.getRecord().set(i, arrData);
                break;
            case MAP:
            case STRUCT:
                Map<String, Object> mapData = isNull(data) ? null : JsonUtils.toObject(data, Map.class);
                row.getRecord().set(i, mapData);
                break;
            default:
                throw new IllegalArgumentException("Unsupported data type: " + dataType.getName());
        }
    }

    private boolean isNull(String data) {
        return data == null || data.equals(NULL_VALUE_FOR_CSV);
    }

    @Override
    public void close() throws IOException {
        if (closed) {
            throw new IllegalStateException("The reader state is closed");
        }

        closed = true;
    }

    @Override
    public boolean hasNext() {
        return (iterator != null && iterator.hasNext()) || (nextIterIdx < totalIter);
    }
}
