package com.huaweicloud.dli.io.read;

import com.huaweicloud.dli.io.common.ArrayType;
import com.huaweicloud.dli.io.common.Column;
import com.huaweicloud.dli.io.common.DataType;
import com.huaweicloud.dli.io.common.MapType;
import com.huaweicloud.dli.io.common.PrimitiveType;
import com.huaweicloud.dli.io.common.StructType;
import com.huaweicloud.sdk.core.utils.JsonUtils;

import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.stream.Collectors;

public class SchemaUtils {
    private static final String DECIMAL_TYPE_REGEX = "(?i)^decimal\\((\\d+)(,(\\d+))?\\)$";

    public static List<Column> getSchemaFromJson(String jsonStr) {
        ResultSchema resultSchema = JsonUtils.toObject(jsonStr, ResultSchema.class);
        return resultSchema.getFields().stream()
            .map(resultField -> new Column(resultField.getName(), getDataType(resultField.getType())))
            .collect(Collectors.toList());
    }

    @SuppressWarnings({"unchecked"})
    private static DataType getDataType(Object inType) {
        if (inType instanceof String) {
            String ptStr = ((String) inType).toLowerCase(Locale.ROOT);
            DataType.TypeName typeName = DataType.TypeName.fromName(ptStr);
            switch (typeName) {
                case BOOLEAN:
                case TINYINT:
                case SMALLINT:
                case INT:
                case BIGINT:
                case FLOAT:
                case DOUBLE:
                case DATE:
                case TIMESTAMP:
                case STRING:
                case BINARY:
                case VOID:
                    return new PrimitiveType(typeName);
                case DECIMAL:
                    ptStr = ptStr.replaceAll(" ", "");
                    String precision = ptStr.replaceAll(DECIMAL_TYPE_REGEX, "$1");
                    String scale = ptStr.replaceAll(DECIMAL_TYPE_REGEX, "$3");
                    if (precision.isEmpty()) {
                        return new PrimitiveType(DataType.TypeName.DECIMAL);
                    }
                    if (scale.isEmpty()) {
                        return new PrimitiveType(DataType.TypeName.DECIMAL, Arrays.asList(precision, "18"));
                    }
                    return new PrimitiveType(DataType.TypeName.DECIMAL, Arrays.asList(precision, scale));
                default:
                    throw new IllegalArgumentException("Unsupported data type: " + ptStr);
            }
        }
        if (inType instanceof Map) {
            Map<String, Object> ctMap = (Map<String, Object>) inType;
            String ctName = ((String) ctMap.get("type")).toLowerCase(Locale.ROOT);
            if (ctName.equals("array")) {
                return new ArrayType(getDataType(ctMap.get("elementType")));
            }
            if (ctName.equals("map")) {
                return new MapType(getDataType(ctMap.get("keyType")), getDataType(ctMap.get("valueType")));
            }
            if (ctName.equals("struct")) {
                return new StructType(getSchemaFromJson(JsonUtils.toJSON(inType)));
            }
        }
        throw new IllegalArgumentException("Unsupported data type");
    }
}
