package com.huaweicloud.dli.io.write;

import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.huaweicloud.dli.io.common.Column;
import com.huaweicloud.dli.io.common.DataType;
import com.huaweicloud.dli.io.common.ObsProxy;
import com.huaweicloud.dli.io.common.Row;
import com.huaweicloud.sdk.core.utils.JsonUtils;
import com.obs.services.model.AppendObjectResult;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.UUID;

import lombok.extern.slf4j.Slf4j;

@Slf4j
public class JsonObsWriter implements Writer {
    private static final ThreadLocal<DateFormat> DATE_FORMAT =
        ThreadLocal.withInitial(() -> new SimpleDateFormat("yyyy-MM-dd"));

    private static final ThreadLocal<DateFormat> TIMESTAMP_FORMAT =
        ThreadLocal.withInitial(() -> new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSSZZ"));

    private static final int DATA_BLOCK_SIZE = 8_388_608; // 8MB

    private static final long MAX_SIZE_PER_FILE = 134_217_728; // 128MB

    private final ObsProxy obsProxy;

    private final String bucketName;

    private final String objectPrefix;

    private final ArrayList<DataBlock> dataBlocks = new ArrayList<>();

    private DataBlock dataBlock = null;

    private AppendObjectResult appendObjResult;

    private boolean closed = false;

    public JsonObsWriter(ObsProxy obsProxy, String dataPath) {
        this.obsProxy = obsProxy;
        this.bucketName = obsProxy.getBucketName(dataPath);
        this.objectPrefix = obsProxy.getObjectPrefix(dataPath);
    }

    @Override
    public void write(Row row) throws IOException {
        if (closed) {
            throw new IllegalStateException("The writer state is closed");
        }

        String rowStr = toJsonString(row);
        byte[] rowByteArr = rowStr.getBytes(StandardCharsets.UTF_8);

        if (rowByteArr.length > DATA_BLOCK_SIZE) {
            log.info("create new data block as row length: {} exceed data block size: {}",
                rowByteArr.length, DATA_BLOCK_SIZE);
            DataBlock dataBlockInfo = new DataBlock(rowByteArr.length);
            dataBlockInfo.appendData(rowByteArr);
            this.dataBlocks.add(dataBlockInfo);
            flush();
            return;
        }

        if (this.dataBlock == null) {
            log.info("create new data block as data block not init");
            this.dataBlock = new DataBlock(DATA_BLOCK_SIZE);
            this.dataBlocks.add(this.dataBlock);
        }

        try {
            this.dataBlock.appendData(rowByteArr);
        } catch (NegativeArraySizeException e) {
            log.info("create new data block as current data block is overflows");
            this.dataBlock = new DataBlock(DATA_BLOCK_SIZE);
            this.dataBlocks.add(this.dataBlock);
            this.dataBlock.appendData(rowByteArr);
            flush();
        }
    }

    private String toJsonString(Row row) {
        JsonObject jsonObj = new JsonObject();
        List<Column> schema = row.getSchema();
        for (int i = 0; i < schema.size(); i++) {
            // for static partition column don't write data
            if (schema.get(i).isPartitionColumn()) {
                continue;
            }
            String colName = schema.get(i).getName();
            DataType dataType = schema.get(i).getType();
            DataType.TypeName typeName = DataType.TypeName.fromName(dataType.getName());
            switch (typeName) {
                case BOOLEAN:
                    jsonObj.addProperty(colName, row.getBoolean(i));
                    break;
                case TINYINT:
                    jsonObj.addProperty(colName, row.getTinyint(i));
                    break;
                case SMALLINT:
                    jsonObj.addProperty(colName, row.getSmallint(i));
                    break;
                case INT:
                    jsonObj.addProperty(colName, row.getInt(i));
                    break;
                case BIGINT:
                    jsonObj.addProperty(colName, row.getBigint(i));
                    break;
                case FLOAT:
                    jsonObj.addProperty(colName, row.getFloat(i));
                    break;
                case DOUBLE:
                    jsonObj.addProperty(colName, row.getDouble(i));
                    break;
                case DECIMAL:
                    jsonObj.addProperty(colName, row.getDecimal(i));
                    break;
                case DATE:
                    Date dtV = row.getDate(i);
                    jsonObj.addProperty(colName, dtV == null ? null : DATE_FORMAT.get().format(dtV));
                    break;
                case TIMESTAMP:
                    Timestamp tsV = row.getTimestamp(i);
                    jsonObj.addProperty(colName, tsV == null ? null : TIMESTAMP_FORMAT.get().format(tsV));
                    break;
                case STRING:
                    jsonObj.addProperty(colName, row.getString(i));
                    break;
                case ARRAY:
                case MAP:
                case STRUCT:
                    JsonElement elem = JsonParser.parseString(JsonUtils.toJSON(row.getRecord().get(i)));
                    jsonObj.add(colName, elem);
                    break;
                default:
                    throw new IllegalArgumentException("Unsupported data type: " + dataType.getName());
            }
        }
        return jsonObj + "\n";
    }

    @Override
    public void flush() throws IOException {
        if (closed) {
            throw new IllegalStateException("The writer state is closed");
        }
        if (dataBlocks.size() > 1) {
            writeDataBlocks();
        } else {
            log.info("no need to flush data as not full");
        }
    }

    @Override
    public void close() throws IOException {
        if (closed) {
            throw new IllegalStateException("The writer state is closed");
        }

        writeDataBlocks();
        closed = true;
    }

    private void writeDataBlocks() {
        if (dataBlocks.isEmpty()) {
            log.info("there is not data blocks to write");
            return;
        }

        Iterator<DataBlock> iterator = dataBlocks.iterator();
        InputStream inputStream = new IterableInputStream<>(iterator, DataBlock::getData, DataBlock::getLength);
        if (appendObjResult == null) {
            String objKey = objectPrefix + String.format("/part-%s.json", UUID.randomUUID());
            log.info("flush {} data blocks to {}", dataBlocks.size(), objKey);
            appendObjResult = obsProxy.appendObject(bucketName, objKey, inputStream);
        } else {
            log.info("append {} data blocks to {}", dataBlocks.size(), appendObjResult.getObjectKey());
            appendObjResult = obsProxy.appendObject(
                bucketName, appendObjResult.getObjectKey(), appendObjResult.getNextPosition(), inputStream);
            if (appendObjResult.getNextPosition() > MAX_SIZE_PER_FILE) {
                log.info("current file size has reached the excepted value");
                appendObjResult = null;
            }
        }

        this.dataBlock = null;
        dataBlocks.clear();
    }
}
