package com.huaweicloud.dli.util;

import com.huaweicloud.dli.io.common.Column;
import com.huaweicloud.dli.io.common.Row;
import com.huaweicloud.dli.io.read.DownloadJob;
import com.huaweicloud.dli.io.read.Reader;
import com.huaweicloud.dli.io.write.UploadJob;
import com.huaweicloud.dli.io.write.Writer;

import java.io.IOException;
import java.util.List;
import java.util.function.Consumer;

import lombok.extern.slf4j.Slf4j;

@Slf4j
public class TunnelUtils {
    private static final int FLUSH_THRESHOLD = 100_000;

    private static final long MILLIS_PER_SECOND = 1000L;

    public static void writeData(UploadJob uploadJob, List<Column> schema, int recordNum, Consumer<Row> updateRow)
        throws IOException {
        Writer writer = uploadJob.createWriter();
        int count = 0;
        for (int i = 0; i < recordNum; i++) {
            count++;
            Row row = new Row(schema);
            updateRow.accept(row);
            writer.write(row);
            if (count % FLUSH_THRESHOLD == 0) {
                writer.flush();
            }
        }
        writer.flush();
        writer.close();
    }

    public static void waitUploadFinish(UploadJob uploadJob, boolean shouldCommit)
        throws IOException {
        if (shouldCommit) {
            uploadJob.beginCommit();
        }
        String status = uploadJob.getInnerJobStatus();
        while (notTerminatedStatus(status)) {
            sleepOneSec();
            status = uploadJob.getInnerJobStatus();
        }
        if (isFailedStatus(status)) {
            throw new IOException("Upload job failed");
        }
    }

    public static void readData(Reader reader, boolean shouldOpen, Consumer<Row> handleRow)
        throws IOException {
        if (shouldOpen) {
            reader.openReader();
        }
        while (reader.hasNext()) {
            Row row = reader.read();
            handleRow.accept(row);
        }
    }

    public static void waitDownloadFinish(DownloadJob downloadJob)
        throws IOException {
        String status = downloadJob.getInnerJobStatus();
        while (notTerminatedStatus(status)) {
            sleepOneSec();
            status = downloadJob.getInnerJobStatus();
        }
        if (isFailedStatus(status)) {
            throw new IOException("Download job failed");
        }
    }

    private static void sleepOneSec() {
        try {
            Thread.sleep(MILLIS_PER_SECOND);
        } catch (InterruptedException e) {
            log.error("Unexpected interrupted while sleep", e);
        }
    }

    private static boolean notTerminatedStatus(String status) {
        return !status.equals("FINISHED") && !isFailedStatus(status);
    }

    private static boolean isFailedStatus(String status) {
        return status.equals("FAILED") || status.equals("CANCELLED");
    }
}
